import io
import math
import time
from enum import Enum
from threading import Thread
from typing import Union, Iterator, BinaryIO, Optional, Callable
from functools import partial

from .constants import MAXIMUM_ALLOWED_SIZE, TRANSFER_BUF_SIZE_B
from .exceptions import TransferAborted


class WorkerThread(Thread):
    """
    Worker thread performing a generic `job`.

    Args:
        job:    A callable object expecting this `WorkerThread` object as first and sole argument.
        args:   Other positional arguments for the underlying :py:class:`threading.Thread` object.
        kwargs: Other named arguments for the underlying :py:class:`threading.Thread` object.
    """

    def __init__(self, job: Callable, *args, **kwargs):
        super(WorkerThread, self).__init__(*args, **kwargs)
        self._job = job
        self._is_shutdown = False
        setattr(self, "run", partial(self._job, worker=self))

    @property
    def is_shutdown(self) -> bool:
        """
        Whether the worker is interrupted.
        """
        return self._is_shutdown

    def shutdown(self):
        """
        Interrupts the worker.
        """
        self._is_shutdown = True


class TransferStatus(Enum):
    """
    Models the status of a transfer operation.

    UNKNOWN:
        Unknown status.

    READY:
        Transfer is ready but has not started yet.

    ACTIVE:
        Transfer is currently active. Data is flowing.

    STOPPED:
        The transfer was stopped before the end.

    FINISHED:
        The transfer successfully finished.

    ERROR:
        An error occurred and the transfer was interrupted.
    """

    UNKNOWN = 0
    READY = 1
    ACTIVE = 10
    STOPPED = 20
    FINISHED = 30
    ERROR = 90


class TransferProgress:
    """
    Models the progress of a transfer operation.

    Args:
        total:          Total number of bytes to transfer.
        transferred:    Number of bytes transferred so far.
        part:           Number of the part being transferred.
        parts:          Total number of parts to transfer.
    """

    def __init__(self, total: int, transferred: int = 0, part: int = 1, parts: int = 1):
        self._total = total
        self._transferred = transferred
        self._part = part
        self._parts = parts
        self._speed = 0
        self._percentage = 0
        self._last_update_time = None
        self._callbacks = set()

    @property
    def total(self) -> int:
        """
        Total number of bytes to transfer.
        """
        return self._total

    @property
    def transferred(self) -> int:
        """
        Number of bytes transferred so far.
        """
        return self._transferred

    @property
    def speed(self) -> float:
        """
        Current transfer speed in `bytes/sec`.
        """
        return self._speed

    @property
    def percentage(self) -> float:
        """
        Current progress in percentage.
        """
        return self._percentage

    @property
    def part(self) -> int:
        """
        Number of the part being transferred.
        """
        return self._part

    @property
    def parts(self) -> int:
        """
        Total number of parts to transfer.
        """
        return self._parts

    def register_callback(self, callback: Callable):
        """
        Registers a callback function that will be called every time an update to the progress
        is available.

        Args:
            callback:   A callable object expecting this `TransferProgress` object as first and
                        sole argument.
        """
        self._callbacks.add(callback)

    def unregister_callback(self, callback: Callable):
        """
        Unregisters a callback function previously registered.

        Args:
            callback:   A callable object.
        """
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def update(self, total: int = None, transferred: int = None, part: int = None, parts: int = None):
        """
        Updates the internal state of the object.

        Args:
            total:          (Optional) Total number of bytes to transfer.
            transferred:    (Optional) Number of bytes transferred so far.
            part:           (Optional) Number of the part being transferred.
            parts:          (Optional) Total number of parts to transfer.
        """
        if total is not None:
            self._total = total
        # ---
        if transferred is not None:
            new_data_len = transferred - self._transferred
            # compute speed
            if new_data_len > 0:
                now = time.time()
                if self._last_update_time is not None:
                    self._speed = new_data_len / (now - self._last_update_time)
                self._last_update_time = now
            # compute percentage
            self._percentage = math.floor(100 * transferred / self._total) if self._total else 0
            # update transferred
            self._transferred = transferred
        # ---
        if part is not None:
            self._part = part
        # ---
        if parts is not None:
            self._parts = parts
        # fire a new update event
        self._fire()

    def _fire(self):
        """
        Fires an update event to the registered callbacks.
        """
        for callback in self._callbacks:
            callback(self)

    def __str__(self):
        """
        Stringifies the object.
        """
        return str(
            {
                "total": self._total,
                "transferred": self._transferred,
                "speed": self._speed,
                "percentage": self._percentage,
                "part": self._part,
                "parts": self._parts,
            }
        )


class TransferHandler:
    """
    Models a transfer operation.

    Args:
        progress:   The progress monitor for this transfer operation.
    """

    def __init__(self, progress: TransferProgress):
        self._progress = progress
        self._status = TransferStatus.UNKNOWN
        self._reason = "(none)"
        self._workers = set()
        self._callbacks = set()
        # register the transfer handler as a progress callback
        self._progress.register_callback(self._fire)

    @property
    def status(self) -> TransferStatus:
        """
        The current status of this transfer operation.
        """
        return self._status

    @status.setter
    def status(self, new_status: TransferStatus):
        """
        Sets new transfer status.

        Args:
            new_status: New status.
        """
        if not isinstance(new_status, TransferStatus):
            raise ValueError(
                f"Expected `new_status` of type `TransferStatus`. " f"Got `{str(type(new_status))}` instead."
            )
        self._status = new_status

    @property
    def progress(self) -> TransferProgress:
        """
        The progress monitor for this transfer operation.
        """
        return self._progress

    @property
    def reason(self) -> str:
        """
        A textual description of why the transfer is in the current status.
        For example, when the status is `ERROR`, this carries an error message.
        """
        return self._reason

    def set_status(self, new_status: TransferStatus, reason: str):
        """
        Sets transfer status and relative reason.

        Args:
            new_status: New status.
            reason:     A description of what triggered this change in status.
        """
        self.status = new_status
        self._reason = reason

    def add_worker(self, worker: WorkerThread):
        """
        Adds a worker to this tranfer operation.

        Args:
            worker: A worker thread object performing a job relative to this transfer operation.
        """
        self._workers.add(worker)

    def register_callback(self, callback: Callable):
        """
        Registers a callback function for updates.

        Args:
            callback:   A callable object expecting this `TransferHandler` object as first and
                        sole argument.
        """
        self._callbacks.add(callback)

    def unregister_callback(self, callback: Callable):
        """
        Unregisters a callback function previously registered.

        Args:
            callback:   A callable object.
        """
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def abort(self, block: bool = False):
        """
        Aborts the transfer operation.

        Args:
            block:  Whether to block until all jobs associates to this tranfer are fully stopped.
        """
        # stop workers
        for worker in self._workers:
            if not worker.is_shutdown:
                worker.shutdown()
        # wait for workers to finish
        if block:
            self.join()

    def join(self):
        """
        Blocks until the operation is completed.
        """
        for worker in self._workers:
            # noinspection PyBroadException
            try:
                worker.join()
            except BaseException:
                pass

    def _fire(self, *_, **__):
        """
        Fires an update event to the registered callbacks.
        """
        for callback in self._callbacks:
            callback(self)

    def __str__(self):
        """
        Stringifies the object.
        """
        return str(self._progress)


class IterableIO:
    """
    A converter object turning a file-like object into an iterator of chunks of bytes.

    Args:
        stream:     Underlying stream to consume.
        bufsize:    Buffer size in number of `bytes`. Each chunk will have at most this size.
    """

    def __init__(self, stream: Union[io.RawIOBase, BinaryIO], bufsize: int = TRANSFER_BUF_SIZE_B):
        self._stream = stream
        self._bufsize = bufsize

    def __iter__(self) -> Iterator[bytes]:
        """
        Iterator of chunks of bytes.
        """
        buf = self._stream.read(self._bufsize)
        while len(buf):
            yield buf
            buf = self._stream.read(self._bufsize)


class RangedStream(io.RawIOBase):
    """
    Masked file-like object that consumes bytes from the region [`seek`, `seek+limit`] of
    the underlying file-like object `stream`.

    Args:
        stream: Underlying file-like object to consume.
        seek:   Position of the first byte to consume from `stream`.
        limit:  Total number of bytes to consume from `stream`.
    """

    def __init__(self, stream: io.RawIOBase, seek: int, limit: int):
        self._stream = stream
        self._seek = seek
        self._transferred = 0
        self._limit = limit
        self._initialized = False

    def close(self):
        """
        Closes the stream.
        """
        return

    def read(self, size: int = ...) -> Optional[bytes]:
        """
        Reads a chunk of bytes of size at most `size`.

        Args:
            size:   Maximum size of bytes to read at once.

        Returns:
            bytes:  Chunk of bytes.
        """
        if not self._initialized:
            self._stream.seek(self._seek)
            self._initialized = True
        size = min(size, self._limit - self._transferred)
        chunk = self._stream.read(size)
        self._transferred += len(chunk)
        return chunk


class MultipartBytesIO:
    """
    Splitter of large stream of bytes.

    Args:
        stream:     Underlying file-like object.
        length:     Length of the stream.
        part_size:  Size of each block in the partition.
    """

    def __init__(self, stream: io.RawIOBase, length: int, part_size: int = MAXIMUM_ALLOWED_SIZE):
        self._stream = stream
        self._stream_length = length
        self._part_size = part_size
        self._start = 0

    def __iter__(self) -> Iterator[RangedStream]:
        """
        Iterator of `RangedStream` object.
        """
        for i in range(self.number_of_parts()):
            cursor = i * self._part_size
            part_length = min(self._stream_length - cursor, self._part_size)
            yield part_length, RangedStream(self._stream, cursor, self._part_size)

    def number_of_parts(self) -> int:
        """
        Number of parts (partition blocks).

        Returns:
            int:    Number of parts produced by this splitter.
        """
        return int(math.ceil(self._stream_length / self._part_size))


class MonitoredIOIterator:
    """
    Monitored file-like object.
    Consumes an iterator of bytes and updates a TransferProgress object.

    Args:
        progress:   Instance of :py:class:`TransferProgress` to update.
        iterator:   Iterator of bytes to consume.
        worker:     Instance of :py:class:`WorkerThread` performing the transfer job.
    """

    def __init__(
        self,
        progress: TransferProgress,
        iterator: Union[None, Iterator[bytes]] = None,
        worker: Union[None, WorkerThread] = None,
    ):
        self._progress = progress
        self._iterator = iterator
        self._worker = worker
        self._transferred_bytes = 0
        self._last_time = None

    def set_iterator(self, iterator: Iterator[bytes]):
        """
        Updates the underlying iterator of bytes to consume data from.

        Args:
            iterator:   Iterator of bytes to consume.
        """
        self._iterator = iterator

    def set_worker(self, worker: WorkerThread):
        """
        Sets the current worker thread.

        Args:
            worker: Worker thread.
        """
        self._worker = worker

    def __iter__(self):
        """
        Activates a proxy iterator
        """
        return self

    def __next__(self) -> bytes:
        """
        Next element from the iterator.

        Returns:
            bytes:  Chunk of bytes from the underlying iterator.

        Raises:
            StopIteration:      End of the iterator reached.
            TransferAborted:    The worker was interrupted.
        """
        if self._iterator is None:
            raise StopIteration()
        if self._worker.is_shutdown:
            raise TransferAborted()
        # ---
        data = next(self._iterator)
        self._transferred_bytes += len(data)
        # update progress handler
        self._progress.update(transferred=self._transferred_bytes)
        # update time
        self._last_time = time.time()
        # yield
        return data



PUBLIC_STORAGE_URL = "https://duckietown-{bucket}-storage.s3.amazonaws.com/{object}"
BUCKET_NAME = "duckietown-{name}-storage"
MAXIMUM_ALLOWED_SIZE = 5368709120
DATA_API_VERSION = "v1"
DATA_API_URL = "https://data.duckietown.org/%s/{action}/{bucket}/{object}" % DATA_API_VERSION
TRANSFER_BUF_SIZE_B = 1024 ** 2



__version__ = "0.1.9"

from .logging import logger
from .client import DataClient
from .storage import Storage
from .utils import TransferStatus, TransferProgress, TransferHandler
from .exceptions import APIError, TransferError, ConfigurationError
from dt_authentication import InvalidToken

__all__ = [
    "logger",
    "DataClient",
    "Storage",
    "TransferStatus",
    "TransferProgress",
    "TransferHandler",
    "APIError",
    "TransferError",
    "ConfigurationError",
    "InvalidToken",
]



from typing import Union

from .logging import logger
from .api import DataAPI
from .exceptions import ConfigurationError
from .storage import Storage, UserStorage


class DataClient(object):
    """
    Provides an interface to the Duckietown Cloud Storage Service (DCSS).

    Args:
        token (:obj:`str`):         your secret Duckietown Token

    Raises:
        dt_authentication.InvalidToken: The given token is not valid.

    """

    def __init__(self, token: str = None):
        self._api = DataAPI(token)

    @property
    def api(self):
        """
        The low-level Data API client.
        """
        return self._api

    def storage(self, name: str, impersonate: Union[None, int] = None) -> Storage:
        """
        Creates a :py:class:`dt_data_api.Storage` that interfaces to a specific storage
        space among those available on the DCSS.

        Args:
            name (:obj:`str`):          Name of the storage space.
            impersonate (:obj:`int`):   (Optional) ID of the user to impersonate. Only valid
                                        when ``name = "user"``.

        """
        name = name.strip()
        # handle special case of `user` storage space
        if name == "user":
            if self._api.uid is None:
                raise ConfigurationError("The 'user' storage space can only be created from an "
                                         "authenticated client. Please, pass a Duckietown token "
                                         "while creating the 'DataClient' object.")
            return UserStorage(self.api, "user", impersonate=impersonate)
        # warn the user when `impersonate` is not used properly
        if impersonate is not None:
            logger.warning("The argument `impersonate` is taken into account only when accessing "
                           "the 'user' storage space. It will be ignored.")
        # any other DCSS storage unit
        return Storage(self.api, name)



class ConfigurationError(BaseException):
    """
    Client configuration error.
    """

    pass


class APIError(BaseException):
    """
    Error while talking to the RESTful Data API endpoints.
    """

    pass


class TransferError(BaseException):
    """
    Error while transferring data to or from a storage space on the DCSS.
    """

    pass


class TransferAborted(BaseException):
    """
    Internal Use Only.
    Used to force `requests` to interrupt a streaming of data to the DCSS.
    """

    pass



import requests

from typing import Dict

from dt_authentication import DuckietownToken
from .constants import DATA_API_URL
from .exceptions import APIError


class DataAPI(object):
    """
    Provides an interface to the RESTful Data API service. The RESTful Data API service
    generates signed URLs that allow you to perform operations on files on the DCSS.

    You should not use this class yourself, the DataClient will do it for you.

    Args:
        token (:obj:`str`): you secret Duckietown Token

    Raises:
        dt_authentication.InvalidToken: The given token is not valid.
    """

    def __init__(self, token: str):
        self._uid = None
        # validate the token
        if token is not None:
            self._uid = DuckietownToken.from_string(token).uid
        # store the raw token
        self._token = token

    @property
    def uid(self) -> int:
        """ The user ID corresponding to the given token """
        return self._uid

    @property
    def token(self) -> str:
        """ The given token """
        return self._token

    def authorize_request(self, action: str, bucket: str, obj: str, headers: Dict[str, str] = None):
        """
        Authorizes the request to perform a given ``action`` on a given object ``obj``
        in the storage space ``bucket``.

        Args:
            action (:obj:`str`):        Action to perform on the object. Allowed values are listed
                                        `here <https://docs.aws.amazon.com/AmazonS3/latest/API/
                                        API_Operations_Amazon_Simple_Storage_Service.html>`_.

            bucket (:obj:`str`):        Name of the target storage space (e.g., ``public``).

            obj (:obj:`str`):           Path to the file to perform the action on.

            headers (:obj:`dict`):      Dictionary containing extra headers that will be sent
                                        together with the action request once the signed URL is
                                        generated.

        Raises:
            dt_data_api.APIError:       An error occurs while communicating with the DCSS.
        """
        api_url = DATA_API_URL.format(action=action, bucket=bucket, object=obj)
        api_headers = {"X-Duckietown-Token": self._token}
        if headers is not None:
            api_headers.update(headers)
        # request authorization
        res = requests.get(api_url, headers=api_headers)
        if res.status_code != 200:
            raise APIError(f"API Error: Code: {res.status_code} Message: {res.text}")
        # parse answer
        answer = res.json()
        if answer["code"] != 200:
            raise APIError(f'API Error: Code: {answer["code"]} Message: {answer["message"]}')
        # get signed url
        return answer["data"]["url"]



import os
import io
import requests
from bs4 import BeautifulSoup

from typing import Union, BinaryIO, Dict, List

from dt_authentication import DuckietownToken

from . import logger
from .api import DataAPI
from .utils import (
    IterableIO,
    MonitoredIOIterator,
    MultipartBytesIO,
    TransferProgress,
    WorkerThread,
    TransferHandler,
    TransferStatus,
)
from .constants import BUCKET_NAME, PUBLIC_STORAGE_URL, TRANSFER_BUF_SIZE_B
from .exceptions import TransferError, TransferAborted, APIError


class Storage(object):
    """
    Provides an interface to a storage space on the DCSS.

    .. warning::
        You should not create instances of this class yourself. Use the method
        :py:meth:`dt_data_api.DataClient.storage` instead.

    Args:
        api:    An instance of :py:class:`dt_data_api.api.DataAPI` used to communicate with the
                RESTful Data API.
        name:   Name of the storage space to connect to.
    """

    def __init__(self, api: DataAPI, name: str):
        self._api = api
        self._name = name
        self._full_name = BUCKET_NAME.format(name=name)

    @property
    def api(self) -> DataAPI:
        """ The low-level API object used to communicate with the DCSS """
        return self._api

    def list_objects(self, prefix: str) -> List[str]:
        """
        Lists objects starting with a given prefix.

        Args:
            prefix:         The path prefix to start listing from.

        Returns:
            list[str]:      A list of keys identifying the objects.

        Raises:
            dt_data_api.TransferError:  An error occurs while transferring the data from the DCSS.
            dt_data_api.APIError:       An error occurs while communicating with the DCSS.

        """
        prefix = self._sanitize_remote_path(prefix)
        # authorize request
        self._check_token(f"Storage[{self._name}].list_objects_v2(...)")
        url = self._api.authorize_request("list_objects_v2", self._full_name, prefix)
        # send request
        try:
            res = requests.get(url)
        except requests.exceptions.ConnectionError as e:
            raise TransferError(e)
        # parse output
        items = []
        soup = BeautifulSoup(res.text, 'xml')
        for item in soup.find_all('Contents'):
            items.append(item.Key.text)
        # ---
        return items

    def head(self, obj: str) -> Dict[str, str]:
        """
        Retrieves metadata about the object `obj`.

        Args:
            obj:    The object to retrieve the metadata for.

        Returns:
            dict[str, str]:     A key-value mapping containing the metadata.

        Raises:
            FileNotFoundError:          The object was not found in the storage space.
            dt_data_api.TransferError:  An error occurs while transferring the data from the DCSS.
            dt_data_api.APIError:       An error occurs while communicating with the DCSS.

        """
        if self._name == "public":
            # anybody can do this
            url = PUBLIC_STORAGE_URL.format(bucket=self._name, object=obj)
        else:
            # you need permission for this, authorize request
            self._check_token(f"Storage[{self._name}].head(...)")
            url = self._api.authorize_request("head_object", self._full_name, obj)
        # send request
        try:
            res = requests.head(url)
        except requests.exceptions.ConnectionError as e:
            raise TransferError(e)
        # check output
        if res.status_code == 404:
            raise FileNotFoundError(f"Object '{obj}' not found")
        # ---
        return dict(res.headers)

    def download(self, source: str, destination: str, force: bool = False) -> TransferHandler:
        """
        Downloads a file from the storage space.

        Args:
            source:         The path to the file to download in the storage space.
            destination:    The local path the file is downloaded to.
            force:          Whether the destination file is overwritten in case it exists.

        Returns:
            TransferHandler:    An handler to the transfer operation.

        Raises:
            ValueError:                 One of the arguments has an illegal value.
            FileNotFoundError:          The object was not found in the storage space.
            dt_data_api.TransferError:  An error occurs while transferring the data from the DCSS.

        """
        source = self._sanitize_remote_path(source)
        parts = self._get_parts(source)
        # check destination
        if os.path.exists(destination):
            if os.path.isdir(destination):
                raise ValueError(f"The path '{destination}' already exists and is a directory.")
            if not force:
                raise ValueError(
                    f"The destination file '{destination}' already exists. Use "
                    f"`force=True` to overwrite it."
                )
        # get parts metadata
        metas = [self.head(part) for part in parts]
        obj_length = sum([int(r["Content-Length"]) for r in metas])
        # create a transfer handler
        progress = TransferProgress(obj_length, parts=len(parts))
        handler = TransferHandler(progress)
        # set status to READY
        handler.set_status(TransferStatus.READY, "Worker created")

        # clean up job
        def clean_up():
            if os.path.exists(destination) and os.path.isfile(destination):
                os.remove(destination)

        # define downloading job
        def job(worker: WorkerThread, *_, **__):
            # set status to ACTIVE
            handler.set_status(TransferStatus.ACTIVE, "Worker started")
            # open destination
            with open(destination, "wb") as fout:
                # download parts
                for i, part in enumerate(parts):
                    # check worker
                    if worker.is_shutdown:
                        logger.debug("Transfer aborted!")
                        # set status to STOPPED
                        handler.set_status(TransferStatus.STOPPED, "Worker was stopped")
                        # clean up partial files
                        clean_up()
                        # tell the server we are done
                        res.close()
                        # get out of here
                        return
                    # ---
                    # update progress
                    progress.update(part=i + 1)
                    # get url to part
                    if self._name == "public":
                        # anybody can do this
                        url = PUBLIC_STORAGE_URL.format(bucket=self._name, object=part)
                    else:
                        # you need permission for this, authorize request
                        self._check_token(f"Storage[{self._name}].download(...)")
                        try:
                            url = self._api.authorize_request("get_object", self._full_name, part)
                        except APIError as e:
                            # set status to ERROR
                            handler.set_status(TransferStatus.ERROR, str(e))
                            return
                    # send request
                    res = requests.get(url, stream=True)
                    # stream content
                    for chunk in res.iter_content(TRANSFER_BUF_SIZE_B):
                        # check worker
                        if worker.is_shutdown:
                            logger.debug("Transfer aborted!")
                            # set status to STOPPED
                            handler.set_status(TransferStatus.STOPPED, "Worker was stopped")
                            # clean up partial files
                            clean_up()
                            # tell the server we are done
                            res.close()
                            # get out of here
                            return
                        # ---
                        fout.write(chunk)
                        # update progress
                        progress.update(transferred=progress.transferred + len(chunk))

        # create a worker
        worker_th = WorkerThread(job)
        # register worker with the handler
        handler.add_worker(worker_th)
        # start the worker
        worker_th.start()
        # return transfer handler
        return handler

    def upload(
        self, source: Union[str, bytes, BinaryIO], destination: str, length: int = None
    ) -> TransferHandler:
        """
        Uploads a file to the storage space.

        Args:
            source:         `str` - The local path of the file to upload.\n
                            `bytes` - Content of the file as `bytes` object.\n
                            `BinaryIO` - A file-like object.
            destination:    The path to the resulting file in the storage space.
            length:         (Optional) Length of the data in bytes. Only needed when `source`
                            is of type `BinaryIO`.

        Returns:
            TransferHandler:    An handler to the transfer operation.

        Raises:
            ValueError:                 One of the arguments has an illegal value.
            dt_data_api.TransferError:  An error occurs while transferring the data to the DCSS.

        """
        if isinstance(source, str):
            file_path = os.path.abspath(source)
            if not os.path.isfile(file_path):
                raise ValueError(f"The file {file_path} does not exist.")
            source = open(file_path, "rb")
            source_len = os.path.getsize(file_path)
        elif isinstance(source, bytes):
            source_len = len(source)
            source = io.BytesIO(source)
        elif isinstance(source, io.RawIOBase):
            if length is None or 0 < length:
                raise ValueError(
                    "When `source` is a file-like object, the stream `length` "
                    "must be explicitly provided (as number of bytes)."
                )
            source_len = length
        else:
            raise ValueError(
                f"Source object must be either a string (file path), a bytes object "
                f"or a binary stream, got {str(type(source))} instead."
            )
        # prepare owner information
        owner = {"id": "0"}
        if self._api.token is not None:
            token = DuckietownToken.from_string(self._api.token)
            owner["id"] = str(token.uid)
        # ---
        # create a multipart handler
        parts = MultipartBytesIO(source, source_len)
        num_parts = parts.number_of_parts()
        # create a transfer progress handler
        progress = TransferProgress(total=source_len, parts=num_parts)
        # create a transfer handler
        handler = TransferHandler(progress)
        # create a monitored iterator
        monitor = MonitoredIOIterator(progress)
        # sanitize destination
        destination = self._sanitize_remote_path(destination)
        # create destination format
        destination_fmt = lambda p: destination + (f".{p:03d}" if num_parts > 1 else "")
        # set status to READY
        handler.set_status(TransferStatus.READY, "Worker created")

        # define uploading job
        def job(worker: WorkerThread, *_, **__):
            # set status to ACTIVE
            handler.set_status(TransferStatus.ACTIVE, "Worker started")
            # iterate over the parts
            for part, (stream_len, stream) in enumerate(parts):
                if worker.is_shutdown:
                    logger.debug("Transfer aborted!")
                    return
                # ---
                dest_part = destination_fmt(part)
                monitor.set_iterator(iter(IterableIO(stream)))
                # update progress
                progress.update(part=part + 1)
                # round up metadata
                metadata = {
                    "x-amz-meta-number-of-parts": str(num_parts),
                    **{f"x-amz-meta-owner-{k}": v for k, v in owner.items()},
                }
                # authorize request
                self._check_token(f"Storage[{self._name}].upload(...)")
                try:
                    url = self._api.authorize_request(
                        "put_object", self._full_name, dest_part, headers=metadata
                    )
                except APIError as e:
                    # set status to ERROR
                    handler.set_status(TransferStatus.ERROR, str(e))
                    return
                # prepare request
                req = requests.Request("PUT", url, data=monitor).prepare()
                # remove header 'Transfer-Encoding'
                del req.headers["Transfer-Encoding"]
                # add header 'Content-Length'
                req.headers["Content-Length"] = stream_len
                # add metadata
                req.headers.update(metadata)
                # send request
                try:
                    # open a session
                    session = requests.Session()
                    # send request through the session
                    res = session.send(req)
                except requests.exceptions.ConnectionError as e:
                    # set status to ERROR
                    handler.set_status(TransferStatus.ERROR, str(e))
                    logger.debug(f"ERROR: {str(e)}")
                    return
                except TransferAborted:
                    # set status to ERROR
                    handler.set_status(TransferStatus.STOPPED, "Worker was stopped")
                    logger.debug("Worker was stopped!")
                    return
                # parse response
                if res.status_code != 200:
                    # set status to ERROR
                    handler.set_status(TransferStatus.ERROR, res.text)
                    logger.debug(f"Transfer Error: Code: {res.status_code} Message: {res.text}")
                    return
                # set status to FINISHED
                handler.set_status(TransferStatus.FINISHED, "Finished")

        # create a worker
        worker_th = WorkerThread(job)
        # register worker with the handler
        handler.add_worker(worker_th)
        # give the worker to the iterator monitor so that the flow of data can be interrupted
        # when the worker is stopped
        monitor.set_worker(worker_th)
        # start the worker
        worker_th.start()
        # return transfer handler
        return handler

    def _sanitize_remote_path(self, path: str) -> str:
        """
        Sanitizes a remote path for this storage space.

        Args:
            path:           `str` - Path to sanitize.

        Returns:
            str:            Sanitized path.
        """
        return path.lstrip("/")

    def _get_parts(self, obj: str):
        """
        Returns a list of parts to download and append together to form the complete file.

        Args:
            obj:    The path to the object in the storage space.

        Raises:
            FileNotFoundError:  The object was not found in the storage space.

        """
        modes = [
            (False, obj, lambda _: obj),
            (True, obj + ".000", lambda p: obj + f".{p:03d}"),
        ]
        for multipart, source, part_name in modes:
            try:
                res = self.head(source)
                parts = int(res.get("x-amz-meta-number-of-parts", "1"))
                return [obj] if not multipart else [part_name(p) for p in range(parts)]
            except FileNotFoundError:
                pass
        raise FileNotFoundError(f"Object '{obj}' not found")

    def _check_token(self, resource: str = None):
        """
        Checks if the token was set inside the DataClient object.

        Args:
            resource:   (Optional)  A brief description of the service needing the token.

        Raises:
            ValueError:     The token was not set (i.e., the client is unathenticated).
        """
        if self._api.token is None:
            resource = "This resource" if not resource else f"The rosource {resource}"
            raise ValueError(
                f"{resource} requires a valid token. Initialize the DataClient "
                f"object with the `token` argument set."
            )


class UserStorage(Storage):
    """
    Provides an interface to a user's private space on the DCSS.

    .. warning::
        You should not create instances of this class yourself. Use the method
        :py:meth:`dt_data_api.DataClient.storage` and pass the argument `"user"` instead.

    Args:
        api:            An instance of DataAPI used to communicate with the RESTful Data API.
        name:           Name of the storage space to connect to.
        impersonate:
    """

    def __init__(self, api: DataAPI, name: str, impersonate: Union[None, int] = None):
        super().__init__(api, name)
        self._impersonate = impersonate

    def _sanitize_remote_path(self, path: str) -> str:
        """
        Sanitizes a remote path for this storage space.

        Args:
            path:           `str` - Path to sanitize.

        Returns:
            str:            Sanitized path.
        """
        path = super(UserStorage, self)._sanitize_remote_path(path)
        uid = self._impersonate or self._api.uid
        user_dir = f"{uid}/"
        if not path.startswith(user_dir):
            path = os.path.join(user_dir, path)
        return path



from setuptools import find_packages, setup

# :==> Fill in your project data here
# The package name is the name on PyPI
# it is not the python module names.
distro = "daffy"
package_name = "dt-data-api-{}".format(distro)
library_webpage = "http://github.com/duckietown/lib-dt-data-api"
maintainer = "Andrea F. Daniele"
maintainer_email = "afdaniele@ttic.edu"
short_description = "Data API client library for Duckietown Cloud Storage"
full_description = """
Data API client library for Duckietown cloud storage.
"""

# Read version from the __init__ file
def get_version_from_source(filename):
    import ast

    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version_from_source("src/dt_data_api/__init__.py")

# read project dependencies
# NO - dependencies.txt is for testing dependencies - EVERYTHING PINNED
# The requirements here must be broad.
# dependencies_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dependencies.txt')
# with open(dependencies_file, 'rt') as fin:
#     dependencies = list(filter(lambda line: not line.startswith('#'), fin.read().splitlines()))

install_requires = ["requests", "beautifulsoup4", "lxml", "dt-authentication-{}".format(distro)]
tests_require = []

# compile description
underline = "=" * (len(package_name) + len(short_description) + 2)
description = """
{name}: {short}
{underline}

{long}
""".format(
    name=package_name, short=short_description, long=full_description, underline=underline
)

console_scripts = []
# setup package
setup(
    name=package_name,
    author=maintainer,
    author_email=maintainer_email,
    url=library_webpage,
    tests_require=tests_require,
    install_requires=install_requires,
    package_dir={"": "src"},
    packages=find_packages("./src"),
    long_description=description,
    version=version,
    entry_points={"console_scripts": console_scripts},
)
